import { Prec } from "@codemirror/state";
import { Awareness, LoroDoc, UndoManager } from "loro-crdt";
import { createCursorLayer, createSelectionLayer, AwarenessPlugin, remoteAwarenessStateField, RemoteAwarenessPlugin, loroCursorTheme, } from "./awareness.js";
import { LoroSyncPluginValue } from "./sync.js";
import { keymap, ViewPlugin } from "@codemirror/view";
import { undoKeyMap, undoManagerStateField, UndoPluginValue } from "./undo.js";
export { undo, redo } from "./undo.js";
export { getTextFromDoc } from "./sync.js";
/**
 * It is used to sync the document with the remote users.
 *
 * @param doc - LoroDoc instance
 * @returns Extension
 */
export const LoroSyncPlugin = (doc) => {
    return ViewPlugin.define((view) => new LoroSyncPluginValue(view, doc));
};
/**
 * LoroAwarenessPlugin is a plugin that adds awareness to the editor.
 * It is used to sync the cursor position and selection of the editor with the remote users.
 *
 * @param doc - LoroDoc instance
 * @param awareness - Awareness instance
 * @param user - User info
 * @param getUserId - Function to get the user id. If not provided, the doc's peerId will be used.
 * @returns Extension[]
 */
export const LoroAwarenessPlugin = (doc, awareness, user, getUserId) => {
    return [
        remoteAwarenessStateField,
        createCursorLayer(),
        createSelectionLayer(),
        ViewPlugin.define((view) => new AwarenessPlugin(view, doc, user, awareness, getUserId)),
        ViewPlugin.define((view) => new RemoteAwarenessPlugin(view, doc, awareness)),
        loroCursorTheme,
    ];
};
/**
 * LoroUndoPlugin is a plugin that adds undo/redo to the editor.
 *
 * @param doc - LoroDoc instance
 * @param undoManager - UndoManager instance
 * @returns Extension[]
 */
export const LoroUndoPlugin = (doc, undoManager) => {
    return [
        undoManagerStateField.init(() => undoManager),
        Prec.high(keymap.of([...undoKeyMap])),
        ViewPlugin.define((view) => new UndoPluginValue(view, doc, undoManager)),
    ];
};
export function LoroExtensions(doc, awareness, undoManager) {
    let extension = [
        ViewPlugin.define((view) => new LoroSyncPluginValue(view, doc))
            .extension,
    ];
    if (undoManager) {
        extension = extension.concat([
            undoManagerStateField.init(() => undoManager),
            Prec.high(keymap.of([...undoKeyMap])),
            ViewPlugin.define((view) => new UndoPluginValue(view, doc, undoManager)).extension,
        ]);
    }
    if (awareness) {
        extension = extension.concat([
            remoteAwarenessStateField,
            createCursorLayer(),
            createSelectionLayer(),
            ViewPlugin.define((view) => new AwarenessPlugin(view, doc, awareness.user, awareness.awareness, awareness.getUserId)),
            ViewPlugin.define((view) => new RemoteAwarenessPlugin(view, doc, awareness.awareness)),
            loroCursorTheme,
        ]);
    }
    return extension;
}
